#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#           Mathlib 2 (sympy powered)         #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

## Cette librairie contient les fonctions de haut niveau non inclues dans sympy.

import sympy
import custom_objects
import math
import internal_functions
import sympy_functions
import intervalles
import pylib

def deg(x):
    if isinstance(x, (list, tuple)):
        x = pylab.array(x)
    return x*180/math.pi

def rad(x):
    if isinstance(x, (list, tuple)):
        x = pylab.array(x)
    return x*math.pi/180


def cbrt(x):
    u"Racine cubique de x."
    return cmp(x, 0)*math.exp(math.log(abs(x))/3)

def root(x, n):
    u"""Racine nime de x.
    
    N'est dfinie pour x ngatif que si n est pair."""
    if n%2:
        return cmp(x, 0)*math.exp(math.log(abs(x))/n)
    return math.exp(math.log(x)/n)

def prod(facteurs):
    return reduce(lambda x,y:x*y, facteurs, 1)


def gcd(a, b):
    u"pgcd de a et de b"
    # algorithme d'Euclide
    a, b = max(abs(a),abs(b)), min(abs(a),abs(b))
    while b:
        a, b = b, a%b
    return a

def lcm(a, b):
    u"ppcm de a et de b"
    return a*b//gcd(a,b)

def pgcd(*termes):
    u"Le plus grand dnominateur commun  un nombre quelconque d'entiers."
    return reduce(lambda x,y:gcd(x,y), termes)

    
def ppcm(*termes):
    u"Le plus petit multiple commun  un nombre quelconque d'entiers."
    return reduce(lambda x,y:lcm(x,y), termes)

#~ #TODO: Rcrire facteurs en utilisant une classe "Produit"
#~ def facteurs(n):
    #~ u"Dcompose n en produit de facteurs premiers."
    #~ n = int(n)
    #~ p = 1 # produit des facteurs
    #~ if n < 0:
        #~ n = abs(n)
        #~ p.append(-1)
    
    #~ while n%2 == 0:
        #~ n = n//2; p *= sympy.var('2')
    
    #~ i = 3
    #~ m = math.sqrt(n)
    #~ while i <= m:
        #~ while n%i == 0:
            #~ n = n//i; m = m//math.sqrt(i)
            #~ p *=sympy.var(str(i))
        #~ i += 2
        
    #~ if n != 1:  p *=sympy.var(str(n))
    
    #~ return p    






def n_premiers(n = 100, maximum = 50000): # securite face aux erreurs de frappe...!
    u"Donne la liste des n premiers nombres premiers."
    liste = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    m = liste[-1] + 2
    if n > maximum:
        raise OverflowError
    while len(liste) < n:
        r = math.sqrt(m)
        for k in liste:
            if k > r: # pas de diviseurs inferieurs  sa racine, donc le nombre est premier
                liste.append(m)
                break
            if not m%k:
                break
        m += 2
    return liste
    
    

    


def frac(x, maximum = 100000):
    u"Convertit un nombre dcimal en fraction." # (par nombre dcimal, il faut entendre float)
    for i in xrange(2, maximum):
        if (int(x*i)-i*x) == 0:
            return int(x*i), i
    for i in xrange(1, maximum):
        if (int(x*10**i)-10**i*x) == 0:
            return int(float(str(float(x*10**i)))), 10**i # le passage par str permet de supprimer les chiffres non significatifs
            
            
def bin(n):
    u"Conversion en binaire."
    s = ""
    while n:
        s = str(n%2) + s
        n //= 2
    return s
    
    
    
    
    
    
    
def custom_latex(expr, inline=True):
    return custom_objects.CustomLatexPrinter(inline).doprint(expr)


def auto_collect(expression):
    u"""Factorise une expression en utilisant sympy.collect, sans prciser manuellement ce par quoi factoriser."""
    if expression.is_Add:
        for terme in expression.args:
            if terme.is_Mul:
                for facteur in terme.args:
                    expression_transformee = sympy.collect(expression, facteur)
                    if expression_transformee.is_Mul:
                        resultat = 1
                        for arg in expression_transformee.args:
                            resultat *= auto_collect(arg)
                        return resultat
        for symbole in internal_functions.syms(expression):
            expression_transformee = sympy.collect(expression, symbole)
            if expression_transformee.is_Mul:
                resultat = 1
                for arg in expression_transformee.args:
                    resultat *= auto_collect(arg)
                return resultat
    return expression

def _Pow2list(expression):
    u"""On dcompose une puissance en liste de facteurs."""
    base, puissance = expression.as_base_exp()
    if puissance.is_integer:
        return int(puissance)*[base]
    elif base == sympy.E:
        coeff = puissance.as_coeff_terms()[0]
        if coeff.is_integer:
            return int(coeff)*[base**(puissance/coeff)]
    return [expression]

def _Mul2list(expression):
    u"""On dcompose un produit en une liste de facteurs ; les puissances sont converties pralablement en produits."""
    liste = []
    if expression.is_Mul:
        for facteur in expression.args:
            liste += _Pow2list(facteur)
        return liste
    return _Pow2list(expression)
        

def auto_collect(expression):
    u"""Factorise une expression en utilisant sympy.collect, sans prciser manuellement ce par quoi factoriser."""
    if expression.is_Add:
        dico = {}
        liste0 = _Mul2list(expression.args[0])
        for elt in liste0:
            dico[elt] = liste0.count(elt)
        for terme in expression.args[1:]:
            liste = _Mul2list(terme)
            for elt in liste0:
                dico[elt] = min(dico[elt], liste.count(elt))
        produit = 1
        for key in dico:
            produit *= key**dico[key]
        return produit*sum(arg/produit for arg in expression.args)
    return expression



def derivee(f):
    return custom_objects.Fonction(f.variables[0], sympy.diff(f.expression, f.variables[0]))


def nul(expression, variable = None):
    u"""Retourne l'ensemble sur lequel une expression  variable relle est nulle."""
    if variable is None:
        variable = internal_functions.extract_var(expression)
    expression = sympy_functions.factor(expression, variable, "R")
    liste_solutions = sympy_functions.solve(expression, variable, ensemble = "R")
    return intervalles.Ensemble_discret(*liste_solutions)
    
    
def ensemble_definition(expression, variable = None):
    if variable is None:
        variable = internal_functions.extract_var(expression)
    expression = sympy_functions.factor(expression, variable, "R")
    if hasattr(expression, "subs"):
        old_variable = variable
        variable = sympy.Symbol("_tmp",real=True)
        expression = expression.subs({old_variable:variable})
    ens_def = intervalles.R
    if hasattr(expression, "is_Pow") and expression.is_Pow:
        base, p = expression.as_base_exp()
        if p.is_integer:
            if p >= 0:
                return ensemble_definition(base, variable)
            else:
                return ensemble_definition(base, variable) - nul(base, variable)
        elif p.is_rational:
            n, d = p.as_numer_denom()
            if abs(n) == 1 and not d.is_even:
                if p >= 0:
                    return ensemble_definition(base, variable)
                else:
                    return ensemble_definition(base, variable) - nul(base, variable)
            else:
                if p >= 0:
                    return positif(base, variable)
                else:
                    return positif(base, variable, strict = True)
        else:
            if p >= 0:
                return positif(base, variable)
            else:
                return positif(base, variable, strict = True)
    if hasattr(expression, "is_Mul") and expression.is_Mul:
        for facteur in expression.args:
            ens_def *= ensemble_definition(facteur, variable)
        return ens_def
    if hasattr(expression, "is_Add") and expression.is_Add:
        for terme in expression.args:
            ens_def *= ensemble_definition(terme, variable)
        return ens_def
    if isinstance(expression, sympy.tan):
        arg = expression.args[0] # -pi/2 < arg < pi/2
        return positif(arg + sympy.pi/2, variable, strict = True)*positif(sympy.pi/2-arg, variable, strict = True)
    elif isinstance(expression, sympy.ln):
        arg = expression.args[0] # 0 < arg
        return positif(arg, variable, strict = True)
    return ens_def
        
        
def positif(expression, variable = None, strict = False):
    u"""Retourne l'ensemble sur lequel une expression  variable relle est positive (resp. strictement positive)."""
    # L'tude du signe se fait dans R, on indique donc  sympy que la variable est relle.
    if variable is None:
        variable = internal_functions.extract_var(expression)
    expression = sympy_functions.factor(expression, variable, "R")
    if hasattr(expression, "subs"):
        old_variable = variable
        variable = sympy.Symbol("_tmp",real=True)
        expression = expression.subs({old_variable:variable})
    ens_def = ensemble_definition(expression, variable)
    if hasattr(expression, "is_Pow") and expression.is_Pow and expression.as_base_exp()[1].is_rational:
        base, p = expression.as_base_exp()
        # Le dnominateur ne doit pas s'annuler :
        if p < 0:
            strict = True
        if p.is_integer and p.is_even:
            if strict:
                return ens_def*(intervalles.R - (positif(base, variable, strict = False) - positif(base, variable, strict = True)))
            else:
                return ens_def
        else:
            return ens_def*positif(base, variable, strict = strict)
    if hasattr(expression, "is_Mul") and expression.is_Mul:
        posit = intervalles.R
        posit_nul = intervalles.R
        for facteur in expression.args:
            # pos : ensemble des valeurs pour lequelles l'expression est positive
            # pos_nul : ensemble des valeurs pour lequelles l'expression est positive ou nulle
            pos = positif(facteur, variable, strict = True)
            pos_nul = positif(facteur, variable, strict = False)
            # posit : les deux sont strictements positifs, ou les deux sont strictements ngatifs
            # posit_nul : les deux sont positifs ou nuls, ou les deux sont ngatifs ou nuls
            posit, posit_nul = (posit*pos + (-posit_nul)*(-pos_nul))*ens_def, (posit_nul*pos_nul + (-posit)*(-pos))*ens_def
             
##            print "resultat", facteur, res
####            if res is NotImplemented:
####                return NotImplemented
##            # le rsultat est positif si les deux facteurs sont positifs, ou si les deux facteurs sont ngatifs:
##            resultat = resultat*res + (-resultat)*(-res)
        if strict:
            return posit
        else:
            return posit_nul
    if hasattr(expression, "is_positive") and expression.is_positive is True: # > 0
#        print "cas 0"
        return ens_def
    if hasattr(expression, "is_negative") and expression.is_negative is True: # < 0
#        print "cas 1"
        return intervalles.vide
    if hasattr(expression, "is_positive") and expression.is_positive is False and strict: # <= 0
#        print "cas 2"
        return intervalles.vide
    if hasattr(expression, "is_negative") and expression.is_negative is False and not strict: # >= 0
#        print "cas 3"
        return ens_def
    if hasattr(expression, "is_zero") and expression.is_zero is True and not strict: # == 0
#        print "cas 4"
        return ens_def
    if isinstance(expression, (int, float, long)):
        if expression > 0 or (expression == 0 and not strict):
            return ens_def
        else:
            return intervalles.vide
    # pas besoin de l'ensemble de dfinition pour les fonctions polynomiales
    if hasattr(expression, "is_polynomial") and expression.is_polynomial():
        oo = sympy.oo
        Intervalle = intervalles.Intervalle
        P = expression.as_poly()
        if P.degree == 1:
            a, b = P.iter_all_coeffs()
            if a > 0:
                return Intervalle(-b/a, +oo, inf_inclus = not strict)
            else: # a<0 (car a != 0)
                return Intervalle(-oo, -b/a, sup_inclus = not strict)
        elif P.degree == 2:
            a, b, c = P.iter_all_coeffs()
            d = b**2 - 4*a*c
            if d > 0:
                x1 = (-b - sympy.sqrt(d))/(2*a)
                x2 = (-b + sympy.sqrt(d))/(2*a)
                x1, x2 = min(x1, x2), max(x1, x2)
                if a > 0:
                    return Intervalle(-oo, x1, sup_inclus = not strict) + Intervalle(x2, +oo, inf_inclus = not strict)
                else: # a<0 (car a != 0)
                    return Intervalle(x1, x2, inf_inclus = not strict, sup_inclus = not strict)
            elif d == 0:
                x0 = -b/(2*a)
                if a > 0:
                    return Intervalle(-oo, x0, sup_inclus  = not strict) + Intervalle(x0, +oo, inf_inclus = not strict)
                else:
                    return Intervalle(x0, x0, sup_inclus  = not strict)
            else: # d < 0
               if a > 0: 
                   return intervalles.R
               else:
                   return intervalles.vide
    raise NotImplementedError
    
    
def resoudre(chaine, variables = (), local_dict = None):
    if local_dict is None:
        evaluer = sympy.sympify
    else:
        def evaluer(expression, local_dict = local_dict):
            return eval(expression, local_dict.globals, local_dict)

    if not variables:
        arguments = chaine.split(',')
        variables = [sympy.Symbol(s.strip()) for s in arguments[1:]]
        if not variables:
            variables = set()
            expressions = pylib.msplit(arguments[0], ('et', 'ou', '>', '<', '=', '!', ')', '(', '*', '/', '-', '+'))
            for expr in expressions:
                if expr.strip():
#                    print expr
                    ev = evaluer(expr)
                    if hasattr(ev, 'atoms'):
                        variables.update(ev.atoms(sympy.Symbol))
#            variables = list(evaluer(arguments[0].replace("et","+").replace("ou","+")
#                            .replace("<", "+").replace(">", "+")
#                            .replace("=", "+").replace("!", "+")).atoms(sympy.Symbol))
        chaine = arguments[0]
        
    if len(variables) > 1:
        return systeme(chaine, local_dict = local_dict)
#    fin = ",".join(arguments[1:])
#    if fin:
#        fin = "," + fin
    chaine = chaine.replace(')et', ') et').replace(')ou', ') ou').replace('et(', 'et (').replace('ou(', 'ou (')
    debut = ''
    while chaine:
        l = [s for s in pylib.split_around_parenthesis(chaine)]
        if len(l) == 3:
            if l[0].strip() == l[2].strip() == '':
                return resoudre(chaine[1:-1], variables = variables, local_dict = local_dict)
            if ' et ' in l[0]:
                retour = chaine.split(' et ', 1)
                retour[0] = debut + retour[0] 
                return resoudre(retour[0], variables = variables, local_dict = local_dict)*resoudre(retour[1], variables = variables, local_dict = local_dict)
            elif ' ou ' in l[0]:
                retour = chaine.split(' ou ', 1)
                retour[0] = debut + retour[0] 
                return resoudre(retour[0], variables = variables, local_dict = local_dict) + resoudre(retour[1], variables = variables, local_dict = local_dict)
            chaine = l[2]
            debut += l[0] + l[1]
        else:
            if ' et ' in chaine:
                retour = chaine.split(' et ', 1)
                retour[0] = debut + retour[0] 
                return resoudre(retour[0], variables = variables, local_dict = local_dict)*resoudre(retour[1], variables = variables, local_dict = local_dict)
            elif ' ou ' in chaine:
                retour = chaine.split(' ou ', 1)
                retour[0] = debut + retour[0] 
                return resoudre(retour[0], variables = variables, local_dict = local_dict) + resoudre(retour[1], variables = variables, local_dict = local_dict)
            else:
                break
    chaine = debut + chaine
    chaine = chaine.replace("==", "=").replace("<>", "!=").replace("=>", ">=").replace("=<", "<=")
            
    if ">=" in chaine:
        gauche, droite = chaine.split(">=")
        return positif(evaluer(gauche + "-(" + droite + ")"), *variables)
    elif "<=" in chaine:
        gauche, droite = chaine.split("<=")
        return positif(evaluer(droite + "-(" + gauche + ")"), *variables)
    if ">" in chaine:
        gauche, droite = chaine.split(">")
        return positif(evaluer(gauche + "-(" + droite + ")"), *variables, **{"strict": True})
    elif "<" in chaine:
        gauche, droite = chaine.split("<")
        return positif(evaluer(droite + "-(" + gauche + ")"), *variables, **{"strict": True})
    elif "!=" in chaine:
        gauche, droite = chaine.split("!=")
        expression = evaluer(gauche + "-(" + droite + ")")
        return ensemble_definition(expression, *variables) - nul(expression, *variables)
    elif "=" in chaine:
        gauche, droite = chaine.split("=")
        expression = evaluer(gauche + "-(" + droite + ")")
        return nul(expression, *variables)
    else:
        raise TypeError, "'" + chaine + "' must be an (in)equation." 


def systeme(chaine, variables = (), local_dict = None):
    chaine = chaine.replace("==", "=")
    if local_dict is None:
        evaluer = sympy.sympify
    else:
        def evaluer(expression, local_dict = local_dict):
            return eval(expression, local_dict.globals, local_dict)
            
    def transformer(eq):
        gauche, droite = eq.split("=")
        return evaluer(gauche + "-(" + droite + ")")
        
    if not variables:
        arguments = chaine.split(',')
        variables = tuple(sympy.Symbol(s.strip()) for s in arguments[1:])
        chaine = arguments[0]
    eqs = tuple(transformer(eq) for eq in chaine.split("et"))
    if not variables:
        variables = set()
        for eq in eqs:
            variables.update(eq.atoms(sympy.Symbol))
    return sympy.solve(eqs, *variables)
